# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import gc
import time
import copy
import datetime
import logging

import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
# from matplotlib import colors
from matplotlib.ticker import FuncFormatter  # (AutoMinorLocator, MultipleLocator, FuncFormatter)

from scipy import signal

from foglamp.common import iprpc, logger, utils

# import ml_eval

# canonical data field names
KEY_DATA1 = 'Channel0'
KEY_DATA2 = 'Channel1'
KEY_DATA3 = 'Channel2'
KEY_DATA4 = 'Channel3'
KEY_TS = 'user_ts'

# very special: if we are passed in a default logger destination use that
if hasattr(logger, 'FOGLAMP_DESTINATION_ENV') and (logger.FOGLAMP_DESTINATION_ENV in os.environ):
    logger.set_default_destination(int(os.environ[logger.FOGLAMP_DESTINATION_ENV]))
else:
    utils.eprint("NO DEST ENV")

_LOGGER = logger.setup(__name__, level=logging.INFO)

LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}


# @utils.add_functions_as_methods(ml_eval.export_service_functions)
class BackendServer(iprpc.InterProcessRPC):

    def __init__(self):
        """ Initialize class internals """
        super().__init__()

        MAX_FRAMES = 10  # max number of frames that can be buffered

        self.logger = _LOGGER

        self.handle = dict()

        # data are input into an in-memory frame
        self.df = pd.DataFrame()
        self.num_rows_processed = 0
        self.num_input_rows = 0
        self.num_frames = 0
        self.frame = None
        self.frames = []
        self.max_frames = MAX_FRAMES

        self.forwardData = False
        self.valid = False

        self.sampling_rate = 8000
        self.cols = [KEY_TS, KEY_DATA1, KEY_DATA2, KEY_DATA3, KEY_DATA4]

        # self.LocalTimeZone = '/'.join(os.path.realpath('/etc/localtime').split('/')[-2:])
        self.LocalTimeZone = 'Etc/UTC'  # foglamp seems to be always running with UTC timestamp, always true?

        self.destDir = os.path.join('/tmp', 'specgram_plots')
        self.specgramFilesMaxAge = None

        self.periodicEventTimestamp = None
        self.eventRepetitionTime = None
        self.eventPreTime = None
        self.eventDuration = None
        self.eventPostTime = None
        self.processingInterval = None
        self.dataCollectionDuration = None
        self.lastEventSearchBaseTime = None
        self.nextEventTimePre = None

        # self.lastProcessedDataTS = pd.Timestamp('2000-01-01 00:00:00.000000+00:00')

    def getNextDataCollectionTime(self, lastEventSearchBaseTime):
        """ Calculates time when data collection is to be done next
                Args:
                    lastEventSearchBaseTime: Time that was used as base when searching for data collection time, last
                    time around
                Returns:
                    nextEventTimePre: Time when data collection is to be done next
        """
        time_diff = (lastEventSearchBaseTime - self.periodicEventTimestamp).total_seconds()  # time since given event TS
        self.logger.debug("time_diff={} sec".format(time_diff))
        self.logger.debug("eventRepetitionTime={} sec".format(self.eventRepetitionTime))
        timeSinceLastEvent = time_diff % self.eventRepetitionTime

        nextEventTimeSec = self.eventRepetitionTime - timeSinceLastEvent
        nextEventTimeSecPre = nextEventTimeSec - self.eventPreTime
        if nextEventTimeSecPre < 0:
            nextEventTimeSecPre = nextEventTimeSecPre + self.eventRepetitionTime
        self.logger.debug("timeSinceLastEvent={} sec, nextEventTimeSec={} sec, nextEventTimeSecPre={} sec"
                          .format(timeSinceLastEvent, nextEventTimeSec, nextEventTimeSecPre))

        nextEventTimePre = lastEventSearchBaseTime + pd.to_timedelta("{} sec".format(nextEventTimeSecPre))
        return nextEventTimePre

    def deleteAgedFiles(self):
        """ Delete spectrogram image folder(s) which are too old as per filter config
                Args:
                Returns:
                    None
        """
        now = time.time()

        for f in os.listdir(self.destDir):
            path = os.path.join(self.destDir, f)
            if os.path.isdir(path) and os.stat(path).st_mtime < (now - self.specgramFilesMaxAge):
                self.logger.info("Directory {} is older than given date, deleting it...".format(f))
                os.system("rm -rf " + path)

    def plugin_init(self, handle):
        """ Initializes class state using passed config
                Args:
                    handle: config dict for the filter
                Returns:
                    handle: config dict for the filter
        """
        self.logger.info('PLUGIN INIT')
        self.logger.setLevel(LOGLEVEL[handle['logLevel']['value']])

        handle = copy.deepcopy(handle)

        # initialize data and common settings
        _sampling_rate = handle['samplingRate']['value']
        if _sampling_rate.endswith('khz'):
            self.sampling_rate = int(_sampling_rate[:-3]) * 1000
        else:
            self.sampling_rate = int(_sampling_rate)

        self.forwardData = (handle['forwardData']['value'] == 'true')
        # self.inference_gap = pd.to_timedelta(handle['evaluationInterval']['value'])

        self.cols = handle['useCols']['value'].split(',')
        
        self.destDir = handle['destDir']['value']
        self.logger.debug("self.specgramFilesMaxAge = {}days".format(handle['specgramFilesMaxAge']['value']))
        self.specgramFilesMaxAge = int(handle['specgramFilesMaxAge']['value'])*24*60*60  # convert from days to seconds
        self.logger.debug("self.specgramFilesMaxAge={} secs".format(self.specgramFilesMaxAge))

        self.periodicEventTimestamp = pd.to_datetime(handle['periodicEventTimestamp']['value'], errors='coerce')
        self.eventRepetitionTime = pd.to_timedelta(handle['eventRepetitionTime']['value'], errors='coerce')
        self.eventPreTime = pd.to_timedelta(handle['eventPreTime']['value'], errors='coerce')
        self.eventDuration = pd.to_timedelta(handle['eventDuration']['value'], errors='coerce')
        self.eventPostTime = pd.to_timedelta(handle['eventPostTime']['value'], errors='coerce')
        self.processingInterval = pd.to_timedelta(handle['processingInterval']['value'], errors='coerce')

        if self.periodicEventTimestamp == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.periodicEventTimestamp))
        else:
            self.periodicEventTimestamp = self.periodicEventTimestamp.tz_localize(tz='UTC')

        if self.eventRepetitionTime == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.eventRepetitionTime))
        else:
            self.eventRepetitionTime = self.eventRepetitionTime.total_seconds()

        if self.eventPreTime == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.eventPreTime))
        else:
            self.eventPreTime = self.eventPreTime.total_seconds()

        if self.eventDuration == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.eventDuration))
        else:
            self.eventDuration = self.eventDuration.total_seconds()

        if self.eventPostTime == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.eventPostTime))
        else:
            self.eventPostTime = self.eventPostTime.total_seconds()

        if self.processingInterval == pd.NaT:
            self.logger.debug("BackendServer: plugin_init(): Unable to process config['{}']".format(
                self.processingInterval))
        else:
            self.processingInterval = self.processingInterval.total_seconds()

        self.dataCollectionDuration = self.eventPreTime + self.eventDuration + self.eventPostTime
        # calculate next collection start time
        self.lastEventSearchBaseTime = pd.Timestamp.now().tz_localize(tz=self.LocalTimeZone).tz_convert('UTC')
        self.nextEventTimePre = self.getNextDataCollectionTime(self.lastEventSearchBaseTime)
        self.logger.debug("Searched with lastEventSearchBaseTime={}, next event pre occurs at {} and data collection "
                          "period is {} secs "
                          .format(self.lastEventSearchBaseTime, self.nextEventTimePre, self.dataCollectionDuration))

        # initialize ML-specific values
        # handle = self.ml_plugin_init(handle)

        self.handle = handle
        self.valid = True
        return handle

    def plugin_shutdown(self):
        """ Shutdown routine for this backend server
                Args:
                Returns:
                    None
        """
        self.logger.warning("BackendServer: SHUTTING DOWN")
        self.valid = False

    def plugin_ingest(self, data):
        """ Processes incoming vibration data; Either accumulates or ignores data as per configured timespec. Also
        creates spectrogram images if accumulation ends
        Args:
            data: Reading data
        Returns:
            small shoot width: width in seconds, if a small shoot is found, else 0
            small shoot start time: start time of small shoot if found, else 0
        """
        self.logger.debug("BackendServer: plugin_ingest")
        # handle = self.handle

        if not self.valid:
            _LOGGER.error("Not processing data since plugin is disabled")
            del data
            return 0, 0

        try:
            # self.logger.debug("srvev: " + str(data))

            if not self.accumulate_data(data):
                del data
                return 0, 0
            else:
                self.logger.debug("self.frame.shape={}".format(self.frame.shape))
                self.logger.debug("self.frame.head=\n{}".format(self.frame.head()))

            self.logger.debug("Generating spectrograms")
            ret_val = self.process_vibration_data()  # return timestamps also where peak is found
            del self.frame
            # self.logger.info(lsos())

            # delete older directories, if present
            self.deleteAgedFiles()

            # report max width result
            width = -1
            small_shoot_start_time = 0
            for val in ret_val:
                if val[0] > width:
                    width = val[0]
                    small_shoot_start_time = val[1]

            del data
            # self.logger.info("Process prepared frame for generating spectrogram and finding energy burst in there.")
            return width, small_shoot_start_time

        except Exception as ex:
            self.logger.exception('Error in BackendServer plugin_ingest {}'.format(str(ex)))
            self.valid = False
            raise

    def accumulate_data(self, data):
        """ Accumulates incoming vibration data as per configured timespec
        Args:
            data: Reading data
        Returns:
            bool: indicating whether accumulation has completed
        """

        if len(data) == 0:
            return False

        # t1 = datetime.datetime.now()

        _input = pd.DataFrame(data[0])
        firstTS = pd.to_datetime(_input['user_ts'][0])  # .tz_localize(tz='UTC')  # TS in first reading
        _input = pd.DataFrame(data[-1])
        lastTS = pd.to_datetime(_input['user_ts'][0])  # .tz_localize(tz='UTC')  # TS in first reading

        collectionStart = self.nextEventTimePre
        collectionStop = self.nextEventTimePre + pd.to_timedelta("{} sec".format(self.dataCollectionDuration))

        self.logger.debug("firstTS={}, lastTS={}, collectionStart={}, collectionStop={}"
                          .format(firstTS, lastTS, collectionStart, collectionStop))

        # Check intersection
        if firstTS > collectionStop or collectionStart > lastTS:
            self.logger.debug("No intersection")

            # it is possible that there is no intersection, but data stream is already past the collectionStop time;
            # in that case, process accumulated data
            if firstTS > collectionStop:
                self.logger.debug("Stream is already past the collectionStop time, processing it...")
            else:
                return False

        if collectionStop > firstTS:
            # find intersection
            fromTs = max(firstTS, collectionStart)
            toTs = min(lastTS, collectionStop)
            self.logger.debug("Intersection from {} to {}".format(fromTs, toTs))

            t1 = datetime.datetime.now()
            _input = pd.DataFrame(data)
            # self.logger.debug("_input.columns={}".format(_input.columns))
            _input.rename(columns={"user_ts": "user_ts_orig"}, inplace=True)
            _input.drop(labels=['id', 'asset', 'ts'], axis=1, inplace=True)
            _input = _input.join(pd.json_normalize(_input['readings']))
            input_cols = _input['readings'][0].keys()
            extra_cols = set(input_cols) - set(self.cols)
            self.logger.debug("input_cols={}, self.cols={}, extra_cols = {}".format(input_cols, self.cols, extra_cols))
            if len(extra_cols):
                _input.drop(labels=list(extra_cols), axis=1, inplace=True)

            _input.index = pd.to_datetime(_input['user_ts_orig'])
            _input = _input[(_input.index >= fromTs) & (_input.index <= toTs)]  # TODO: Maybe move this filtering above
            self.logger.info("Accumulating data from {} to {}".format(_input.index[0], _input.index[-1]))

            lastReceivedTS = _input.index[-1]
            labels_to_remove = ['readings', 'user_ts_orig']
            if 'user_ts' in _input.columns:
                labels_to_remove.append('user_ts')
            _input.drop(labels=labels_to_remove, axis=1, inplace=True)

            t2 = datetime.datetime.now()
            delta = t2 - t1
            delta = int(delta.total_seconds() * 1000 * 1000)  # in usec
            self.logger.debug("df processing took {:.0f} msec for {} readings, time/reading={:.0f} us"
                              .format(delta / 1000, _input.shape[0], delta / _input.shape[0]))

            # self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so processing this data frame"
            #                   .format(self.lastProcessedDataTS, self.inference_gap, currTS))

            self.logger.debug("*** Appending data with TS={}, after data with TS={}"
                              .format(_input.index[0], self.df.index[-1] if self.df.shape[0] else "N.A."))
            _len = _input.shape[0]
            self.df = self.df.append(_input, ignore_index=False)
            self.num_input_rows += _len
            self.num_rows_processed += _len

            del data, _input
            # gc.collect()

            self.logger.debug("### Last TS stored={}, collection stops at {}".format(lastReceivedTS, collectionStop))
            if toTs == lastTS:  # last data row in input stream was selected, so need to wait for more data
                self.logger.debug("### Got {} rows to process, not processing yet".format(self.df.shape[0]))
                self.logger.debug("### Last TS stored={}, collection stops at {}"
                                  .format(self.df.index[-1], collectionStop))
                return False  # don't run model inference if collection has not ended

            self.logger.info("### Data collection finished: from {} to {}".format(self.df.index[0], self.df.index[-1]))

        else:
            self.logger.info("No new data to add, just processing already collected data")

        self.num_frames += 1

        self.frames.append(self.df)
        if len(self.frames) > self.max_frames:
            # keep a rolling list
            f = self.frames.pop(0)
            del f

        self.frame = self.df
        self.df = pd.DataFrame()

        # Collection has stopped; update self.nextEventTimePre
        prevLastEventSearchBaseTime = self.lastEventSearchBaseTime
        self.lastEventSearchBaseTime = self.lastEventSearchBaseTime + pd.to_timedelta("{} sec"
                                                                                      .format(self.processingInterval))
        self.logger.info("lastEventSearchBaseTime updated from {} to {}"
                         .format(prevLastEventSearchBaseTime, self.lastEventSearchBaseTime, ))
        self.nextEventTimePre = self.getNextDataCollectionTime(self.lastEventSearchBaseTime)
        self.logger.info("Next event pre occurs at {} and data collection period is {} secs"
                         .format(self.nextEventTimePre, self.dataCollectionDuration))

        self.logger.debug("### Data collection completed: Processing {} rows...".format(self.frame.shape[0]))

        return True

    def plot_specgram(self, time_series, title, start_ts=None, end_ts=None):
        """ Plots spectrogram image for given time series
        Args:
            time_series: Vibration data for a single channel
            title: Title to use in spectrogram images
            start_ts: Timestamp at beginning of time series
            end_ts: Timestamp at end of time series
        Returns:
            start_x: Offset of beginning of small shoot (energy burst) from start of time series (in secs)
            stop_x: Offset of end of small shoot (energy burst) from start of time series (in secs)
        """
        # nfft = 8 * 1024
        fs = f_s = self.sampling_rate

        if (end_ts - start_ts).total_seconds() < 10.0:
            _nperseg = 256
            _nfft = 1024
            fig_width = 20
            fig_height = 5
        else:
            _nperseg = 256
            _nfft = 256
            fig_width = 26
            fig_height = 5

        freqs, times, Sx = signal.spectrogram(time_series, fs=f_s, window='hanning',
                                              nperseg=_nperseg,
                                              # nperseg=1024//16,  # defaults to 256 if window is str
                                              # noverlap=(1024//16)//8,  # If None, noverlap = nperseg // 8
                                              nfft=_nfft,
                                              detrend='linear',
                                              scaling='spectrum')  # computing the power spectrum

        mn, mx = np.min(Sx), np.max(Sx)
        self.logger.debug("Sx.shape={}, Sx.min/max = {}..{}".format(Sx.shape, mn, mx))
        self.logger.debug("Sx.log-min/max = {}..{}".format(np.log10(mn), np.log10(mx)))
        #     Sx = np.clip(Sx, 5e-8, 0.16)
        #     Sx = np.array(Sx)
        #     Sx[Sx < 5e-10] = 1e-20
        self.logger.debug("fs={}, nfft={}, FFT bin size = {} Hz".format(fs, _nfft, freqs[1]))
        f, ax = plt.subplots(figsize=(fig_width, fig_height))
        psm = ax.pcolormesh(times, freqs, 10 * np.log10(Sx), vmin=-100, vmax=0, shading='gouraud',
                            cmap='rainbow')  # plasma, viridis, rainbow, hsv, coolwarm, Spectral
        f.colorbar(psm, ax=ax)
        ax.axis('tight')
        ax.set_ylabel('Frequency [Hz]')
        ax.set_xlabel('Time [s]')
        #     ax.yaxis.set_major_locator(MultipleLocator(bpfo))
        #     ax.yaxis.set_major_locator(MultipleLocator(rpm/60))
        #     ax.yaxis.set_minor_locator(MultipleLocator(10))
        #     ax.grid(which='both', color='#CC0000', linestyle='--')

        if start_ts:
            def map_x(x, pos):
                """ The two args are the value and tick position """
                return (start_ts + pd.Timedelta("{} sec".format(x))).strftime("%H:%M:%S.%f")

            formatter = FuncFormatter(map_x)
            plt.rcParams.update({'xtick.labelsize': 'large', 'ytick.labelsize': 'large'})
            ax.xaxis.set_major_formatter(formatter)

        y_limit = int(f_s / 2)
        ax.set_ylim(0.0, y_limit)

        ax.set_title(title)
        # self.logger.debug("ax.get_xlim()={}".format(ax.get_xlim()))

        # self.logger.debug("Before log, Sx.shape={}".format(Sx.shape))
        Sx = np.sum(Sx, axis=0)
        self.logger.debug("After log and column wise sum, Sx.shape={}, mean={}, min={}"
                          .format(Sx.shape, np.mean(Sx), np.min(Sx)))

        mn = np.mean(Sx)
        Sx[Sx < mn * 3 / 4] = 0  # mn*2

        #     Sx[Sx>mn*2] = 1

        # def map_x(x, pos):
        #     'The two args are the value and tick position'
        #     t = (end_ts - start_ts) / Sx.shape[0] * x
        #     return (start_ts + t).strftime(".%f")
        # 
        # formatter = FuncFormatter(map_x)

        from scipy.signal import find_peaks, peak_widths
        peaks, _ = find_peaks(Sx, height=mn * 4)

        #     results_half = peak_widths(Sx, peaks, rel_height=0.65)
        #     results_full = peak_widths(Sx, peaks, rel_height=1)
        width, width_height, left_ips, right_ips = peak_widths(Sx, peaks, rel_height=1)
        #     print("results_half[0]={}".format(results_half[0]))
        self.logger.debug("width={}, width_heights={}, left_ips={}, right_ips={}"
                          .format(width, width_height, left_ips, right_ips))
        if len(width) == 0:
            # plt.show()
            today = str(datetime.date.today())
            os.makedirs(os.path.join(self.destDir, today), exist_ok=True)
            plt.savefig(os.path.join(self.destDir, today,
                                     title.replace(" : ", "_").replace(":","_") + '.png'),
                        bbox_inches='tight', facecolor='w')

            plt.cla()  # Clear the current axes
            plt.clf()  # Clear the current figure
            plt.close('all')  # Closes all the figure windows
            plt.close(f)
            # gc.collect()

            self.logger.info("Unable to find any peaks, skipping...")
            return 0, 0

        idx = np.argmax(width)
        width = width[idx]
        width_height = width_height[idx]
        left_ips = left_ips[idx]
        right_ips = right_ips[idx]
        self.logger.debug("max width at index {}: width={}, height={}, range={}..{}"
                          .format(idx, width, width_height, left_ips, right_ips))

        start_x = left_ips / Sx.shape[0] * (end_ts - start_ts).total_seconds()
        stop_x = right_ips / Sx.shape[0] * (end_ts - start_ts).total_seconds()
        self.logger.debug("start_x={} s, stop_x={} s".format(start_x, stop_x))
        # print("Sx.shape[0]={}, segment_len_secs={}, stop_x={} = {} s"
        #       .format(Sx.shape[0], segment_len_secs, stop_x,
        #               stop_x/Sx.shape[0] * (end_ts-start_ts).total_seconds()))
        #     print("start_x={}, stop_x={}".format(start_x, stop_x))
        plt.axvline(x=start_x, linestyle='--', color='black')
        plt.axvline(x=stop_x, linestyle='--', color='black')
        if (end_ts - start_ts).total_seconds() < 10.0:
            today = str(datetime.date.today())
            os.makedirs(os.path.join(self.destDir, today), exist_ok=True)
            plt.savefig(os.path.join(self.destDir, today, title.replace(" : ", "_") + '.png'),
                        bbox_inches='tight', facecolor='w')
            # plt.show()
            plt.cla()  # Clear the current axes
            plt.clf()  # Clear the current figure
            plt.close('all')  # Closes all the figure windows
            plt.close(f)
            # gc.collect()

        idx = np.argmax(Sx)
        t = (end_ts - start_ts) / Sx.shape[0] * idx
        self.logger.debug("idx={}, t={:.0f} ms".format(idx, t.total_seconds() * 1000))
        self.logger.debug("Max value is {} at index {}={}".format(Sx[idx], idx, (start_ts + t).strftime("%H:%M:%S.%f")))
        self.logger.info("small shoot is from offset {} .. {} sec\n".format(start_x, stop_x))

        return start_x, stop_x  # start,stop of small shoot in offset secs

    def process_vibration_data(self):
        """ Plots spectrogram image for accumulated time series
        Args:
        Returns:
            Array of pairs (one for each channel), which contains:
                width: Width of small shoot discharge
                start_time: Start time of small shoot discharge
        """
        df2 = self.frame
        channels = list(df2.columns)
        self.logger.info("----> process_vibration_data(): df2.index[0]={}, df2.index[-1]={}"
                         .format(df2.index[0], df2.index[-1]))
        # self.logger.debug("df2.head()=\n{}".format(df2.head()))

        fs = self.sampling_rate

        start_second = 0
        end_second = df2.shape[0] / fs
        self.logger.debug("start_second={}, end_second={}".format(start_second, end_second))

        ret_val = []

        for channel in channels:
            segment_len_secs = end_second - start_second
            for start_sec in np.arange(0, segment_len_secs, segment_len_secs):
                title = '{} : {} - {} : spectrogram' \
                    .format(channel,  # df2.index[int(fs * (start_sec))].strftime("%Y-%m-%d"),
                            df2.index[int(fs * start_sec)].strftime("%H:%M:%S.%f"),
                            df2.index[int(fs * (start_sec + segment_len_secs)) - 1].strftime("%H:%M:%S.%f"))

                self.logger.debug("title={}".format(title))
                data = df2[channel][int(fs * start_sec):int(fs * (start_sec + segment_len_secs))]

                start_t, stop_t = self.plot_specgram(data, title, start_ts=df2.index[int(fs * start_sec)],
                                                     end_ts=df2.index[int(fs * (start_sec + segment_len_secs)) - 1])

                if start_t == 0 and stop_t == 0:
                    ret_val.append((0, 0))
                    break

                width = stop_t - start_t
                start_ts = df2.index[int(fs * (start_second + start_t))]
                stop_ts = df2.index[int(fs * (start_second + stop_t))]
                self.logger.debug("start_t={:.3f} stop_t={:.3f}, width={:.3f}".format(start_t, stop_t, width))
                self.logger.debug("start_ts={}, stop_ts={}".format(start_ts, stop_ts))

                self.logger.debug("{}: Iter 1: small shoot is from {} .. {}, width={:.0f} msec"
                                  .format(channel, df2.index[int(fs * (start_second + start_t))],
                                          df2.index[int(fs * (start_second + stop_t))], width * 1000))
                start_sec = max(start_second, start_second + start_t - 1)  # 2*width)
                end_sec = min(start_sec + segment_len_secs, start_second + stop_t + 1, end_second)  # 2*width)
                self.logger.debug(
                    "start_sec={}, end_sec={}, segment_len_secs={} -> {}, int(fs*(start_sec+segment_len_secs)-1)={}"
                    .format(start_sec, end_sec, segment_len_secs, end_sec - start_sec,
                            int(fs * (start_sec + segment_len_secs) - 1)))
                segment_len_secs = end_sec - start_sec
                data = df2[channel][int(fs * start_sec):int(fs * (start_sec + segment_len_secs))]
                self.logger.debug("\n\n")

                title = '{} : {} - {} - spectrogram'\
                    .format(channel,
                            # df2.index[int(fs * (start_sec))].strftime("%Y-%m-%d"),
                            df2.index[int(fs * start_sec)].strftime("%H:%M:%S.%f"),
                            df2.index[int(fs * (start_sec + segment_len_secs)) - 1].strftime("%H:%M:%S.%f"))
                start_t, stop_t = self.plot_specgram(data, title, start_ts=df2.index[int(fs * start_sec)],
                                                     end_ts=df2.index[int(fs * (start_sec + segment_len_secs)) - 1])

                if start_t == 0 and stop_t == 0:
                    ret_val.append((0, 0))
                    break

                width = stop_t - start_t
                self.logger.debug("{}: Iter 2: small shoot is from {} .. {}, width={:.0f} msec"
                                  .format(channel, df2.index[int(fs * (start_sec + start_t))],
                                          df2.index[int(fs * (start_sec + stop_t))], width * 1000))
                self.logger.debug("\n====================================\n")

                ret_val.append((width, df2.index[int(fs * (start_sec + start_t))]))
    #     break

        gc.collect()
        # self.logger.info(lsos())

        return ret_val


if __name__ == "__main__":
    rpc = BackendServer()
    rpc.serve()
